import unittest


class TestClassConstsAtRuntime(unittest.TestCase):
    def test_named_class_constants_cannot_be_reassigned(self):
        from constantia.tests.decorators.check_at_runtime.cases.cls.named_constants_cannot_be_reassigned import Example

        with self.assertRaises(ValueError) as context:
            Example()

        self.assertEqual('Reassignment of constant "X" detected on line 4 (X = 8888).', str(context.exception))

    def test_uppercase_class_constants_cannot_be_reassigned(self):
        from constantia.tests.decorators.check_at_runtime.cases.cls.uppercase_constants_cannot_be_reassigned import Example

        with self.assertRaises(ValueError) as context:
            Example()

        self.assertEqual('Reassignment of constant "X" detected on line 4 (X = 8888).', str(context.exception))

    def test_class_constants_cannot_be_reassigned_in_class_method(self):
        from constantia.tests.decorators.check_at_runtime.cases.cls.constants_cannot_be_reassigned_in_class_method import Example

        with self.assertRaises(ValueError) as context:
            e = Example()
            e.change_x()

        self.assertEqual('Reassignment of class constant "X" detected on line 7 (cls.X = 8888).', str(context.exception))

    def test_class_constants_cannot_be_reassigned_in_instance_method(self):
        from constantia.tests.decorators.check_at_runtime.cases.cls.constants_cannot_be_reassigned_in_instance_method import Example

        with self.assertRaises(ValueError) as context:
            e = Example()
            e.change_x()

        self.assertEqual(
            'Reassignment of class constant "X" detected on line 6 (self.__class__.X = 8888).',
            str(context.exception)
        )

    def test_class_constants_cannot_be_reassigned_in_static_method(self):
        from constantia.tests.decorators.check_at_runtime.cases.cls.constants_cannot_be_reassigned_in_static_method import Example

        with self.assertRaises(ValueError) as context:
            Example()

        self.assertEqual(
            'Static reassignment of class constant "X" detected on line 7 (Example.X = 8888).',
            str(context.exception)
        )

    def test_class_constants_valid_assignment(self):
        from constantia.tests.decorators.check_at_runtime.cases.cls.valid_constant_assignment import Example

        Example()



import ast

from constantia.consts.typing import OnType


class ConstAssignmentVisitor(ast.NodeVisitor):
    IMMUTABLE_TYPES = (ast.Constant, ast.Tuple)

    def __init__(self, source_code: str, var_names: OnType = None, klass: type | None = None):
        self.source_code_lines = source_code.splitlines()
        self.var_is_watched = self.__build_var_watch_predicate(var_names=var_names)
        self.cls = klass
        self.constants: set[str] = set()

    def visit(self, node):
        # Assign parent references to all child nodes
        for child in ast.iter_child_nodes(node):
            child.parent = node
        super().visit(node)

    def visit_Assign(self, node):
        for target in node.targets:
            if isinstance(target, ast.Name):
                if not self.var_is_watched(target.id):
                    continue
                # Second assignment
                if target.id in self.constants:
                    code_line = self.__code_line(node)
                    raise ValueError(
                        f'Reassignment of constant "{target.id}" detected on line {node.lineno} ({code_line}).'
                    )
                # Do not allow immutable type assignment
                if not isinstance(node.value, self.IMMUTABLE_TYPES):
                    code_line = self.__code_line(node)
                    raise TypeError(
                        f'Assignment of non-immutable type to constant "{target.id}" detected on line {node.lineno} ({code_line}).'
                    )
                # First assignment of a watched variable
                self.constants.add(target.id)

            # Assignment through class name
            if (
                    isinstance(target, ast.Attribute)
                    and target.attr in self.constants
                    and isinstance(target.value, ast.Name)
                    and target.value.id == self.cls.__name__
            ):
                code_line = self.__code_line(node)
                raise ValueError(
                    f'Static reassignment of class constant "{target.attr}" detected on line {node.lineno} ({code_line}).'
                )

            if (
                    isinstance(target, ast.Attribute)
                    and isinstance(target.value, ast.Attribute)
                    and isinstance(target.value.value, ast.Name)
                    and target.value.value.id == "self"
                    and target.value.attr == "__class__"
                    and target.attr in self.constants
            ):
                code_line = self.__code_line(node)
                raise ValueError(
                    f'Reassignment of class constant "{target.attr}" detected on line {node.lineno} ({code_line}).'
                )

        self.generic_visit(node)

    def _is_self_class_const(self, node):
        """
        Check if the node represents `self.__class__.CONST`.
        """
        return (
                isinstance(node, ast.Attribute)
                and isinstance(node.value, ast.Attribute)
                and isinstance(node.value.value, ast.Name)
                and node.value.value.id == "self"
                and node.value.attr == "__class__"
                and node.attr in self.constants
        )

    def visit_Attribute(self, node):
        if isinstance(node.value, ast.Name) and node.value.id in {'self', 'cls'}:
            if node.attr in self.constants:
                parent = getattr(node, "parent", None)
                if isinstance(parent, ast.Assign):
                    code_line = self.__code_line(node)
                    raise ValueError(
                        f'Reassignment of class constant "{node.attr}" detected on line {node.lineno} ({code_line}).'
                    )

        self.generic_visit(node)

    @staticmethod
    def __build_var_watch_predicate(var_names: OnType):
        if var_names is None:
            return lambda _var: False
        elif isinstance(var_names, list):
            return lambda var: var in var_names
        elif var_names == 'uppercase':
            return lambda var: var.isupper()
        elif callable(var_names):
            return var_names
        else:
            ValueError('Invalid value on on argument. Accepted values are None, a list of strings or a predicate')

    def __code_line(self, node) -> str:
        line = self.source_code_lines[node.lineno - 1]
        return line.strip()



import ast
import inspect
from functools import wraps
from typing import Callable, Literal

from constantia.consts.const_assignment_visitor import ConstAssignmentVisitor
from constantia.consts.typing import OnType


def consts(var_names: OnType = 'uppercase', check_at: Literal["runtime", "import"] = "import"):
    # Handling the no-parameter call
    if callable(var_names):
        obj = var_names
        var_names = 'uppercase'
        check_at: Literal["runtime", "import"] = "import"
        return consts(var_names=var_names, check_at=check_at)(obj)

    def decorator(obj):
        cls = __class_from_obj(obj)

        @wraps(obj)
        def wrapper(*args, **kwargs):
            if check_at == "runtime":
                __assert_constants(func=obj, var_names=var_names, cls=cls)
            return obj(*args, **kwargs)

        if check_at == "import":
            __assert_constants(func=obj, var_names=var_names, cls=cls)

        return wrapper

    return decorator


def __class_from_obj(obj):
    if inspect.isclass(obj):
        return obj
    return None


def __assert_constants(cls: type | None, func: Callable, var_names: OnType = None):
    source_code = inspect.getsource(func)
    tree = ast.parse(source_code)

    visitor = ConstAssignmentVisitor(klass=cls, var_names=var_names, source_code=source_code)
    visitor.visit(tree)



import os
from setuptools import setup, find_packages

root_dir_path = os.path.dirname(os.path.abspath(__file__))

long_description = open(os.path.join(root_dir_path, "README.md")).read()

setup(
    name="constantia",
    version="0.0.5",
    author="Diego J. Romero López",
    author_email="diegojromerolopez@gmail.com",
    description="Enforce constants at import time",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Software Development :: Libraries",
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: MIT License"
    ],
    install_requires=[],
    license="MIT",
    keywords="constants python object",
    url="https://github.com/diegojromerolopez/constantia",
    packages=find_packages(),
    data_files=[],
    include_package_data=True,
    scripts=[]
)



import unittest


class TestFunctionConstsAtRuntime(unittest.TestCase):
    def test_named_function_constants_cannot_be_mutable(self):
        from constantia.tests.decorators.check_at_runtime.cases.function.named_constants_cannot_be_mutable import func

        with self.assertRaises(TypeError) as context:
            func()

        self.assertEqual('Assignment of non-immutable type to constant "x" detected on line 3 (x = [1, 2, 3]).',
                         str(context.exception))

    def test_uppercase_function_constants_cannot_be_mutable(self):
        from constantia.tests.decorators.check_at_runtime.cases.function.uppercase_constants_cannot_be_mutable import func

        with self.assertRaises(TypeError) as context:
            func()

        self.assertEqual('Assignment of non-immutable type to constant "MY_CONSTANT" '
                         'detected on line 3 (MY_CONSTANT = [1, 2, 3]).',
                         str(context.exception))

    def test_function_constants_cannot_be_reassigned(self):
        from constantia.tests.decorators.check_at_runtime.cases.function.constants_cannot_be_reassigned import func

        with self.assertRaises(ValueError) as context:
            func()

        self.assertEqual('Reassignment of constant "x" detected on line 4 (x = 20).', str(context.exception))

    def test_function_constants_valid_assignment(self):
        from constantia.tests.decorators.check_at_runtime.cases.function.valid_constant_assignment import func

        self.assertEqual((1, 2, 3), func())
